
/*****************************************************************************/
/* Copyright (C) 2003-2013, JOANNEUM RESEARCH Forschungs Ges.m.b.H.          */
/* Steyrergasse 17, A-8010 Graz, Austria.                                    */
/*                                                                           */
/*  This file is part of the JRS MPEG-7 Library.                             */
/*  http://mpeg-7.joanneum.at                                                */
/*                                                                           */
/*  JRS MPEG-7 Library is free software: you can redistribute it and/or      */
/*  modify it under the terms of the GNU Lesser General Public License       */
/*  (LGPL) as published by the Free Software Foundation, either version 3    */
/*  of the License, or (at your option) any later version.                   */
/*                                                                           */
/*  JRS MPEG-7 Library is distributed in the hope that it will be useful,    */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the             */
/*  GNU Lesser General Public License for more details.                      */
/*                                                                           */
/*  You should have received a copy of the GNU Lesser General Public License */
/*  along with this library. If not, see <http://www.gnu.org/licenses/>.     */
/*****************************************************************************/


// Based on CoreFactory_cpp.template

#include <stdio.h> // sprintf
#include <xercesc/util/ValueHashTableOf.hpp>
#include <xercesc/util/ValueVectorOf.hpp> // For FactoryVector
#include <xercesc/util/RefVectorOf.hpp> // For NameVector
#include <xercesc/util/XMLString.hpp> // (hashable) GUID type

#include "Mp7JrsFactory.h"
#include "Mp7JrsNode.h"

#include "Mp7JrsFactoryForward.h"

#include <xercesc/util/XMLUniDefs.hpp>
#include "Mp7JrsSchemaLocationMap.h"


int Mp7JrsFactory::initialized = 0;

// The array of registerable factories
typedef ValueVectorOf< Mp7JrsFactory * > Mp7JrsFactoryVector;
static Mp7JrsFactoryVector * FactoryVector;

typedef ValueHashTableOf<unsigned int> Mp7JrsHashTable;
// The hashtable to map a GUID to an index in factory array
static Mp7JrsHashTable * GuidHashTable;
// The hashtable to map a classname to an index in factory array
static Mp7JrsHashTable * NameHashTable;

// The array to map an index to a classname
typedef RefVectorOf< char > Mp7JrsNameVector;
static Mp7JrsNameVector * NameVector;

static Mp7JrsSchemaLocationMap * SchemaLocationMap; // For prefix namespace mapping without XML DOM

// Map a type id to an index within the factory vector
// Returns the ordered index. Index 0 remains unused.
unsigned int Mp7JrsFactory::MapGuidToIndex(SimpleGuid id)
{
	if(!(GuidHashTable && FactoryVector))
		throw Mp7JrsException(MP7JRS_ERROR,"Mp7JrsFactory is not initialised.");
		
	unsigned int i = 0; // Missing entry
	try
	{
	  // FTT TODO in next major release, change id to const XMLCh* instead of SimpleGuid
		i = GuidHashTable->get(X(id)); // FTT Must be X() otherwise we are doomed with extension libs
	}
	catch(...)
	{
		i = (unsigned int)FactoryVector->size(); // unknown type -> to register
	}		
	if(i == 0) // Index 0 is reserved
	{
		i = 1;
	}
	return i;
}



// Register a new factory which can handle a class with that type id
Mp7JrsFactory::Mp7JrsFactory(SimpleGuid id, const char * classname)
{
	if(!FactoryVector) // Dont use ::initialized here!
		throw Mp7JrsException(MP7JRS_ERROR,"Mp7JrsFactory is not initialized.");

	if(!classname) 
		throw Mp7JrsException(MP7JRS_ERROR,"Missing classname, registering failed.");

	// User must decide: overwrite DeleteFactory() or let the base func do its job
	m_ToDelete = false; 
	
	unsigned int i = MapGuidToIndex(id);
	
	if(i >= FactoryVector->size()) // Make room for unregistered factory
	{
		unsigned int oldsize = (unsigned int) FactoryVector->size();
		unsigned needed = i - oldsize + 1;
		unsigned int j, k = oldsize + needed;
		for(j = oldsize; j < k; ++j)
		{
			FactoryVector->addElement((Mp7JrsFactory *)0);
			NameVector->addElement((char *)0);
		}
	}
	
	if(FactoryVector->elementAt(i)) // Replace existing entries
	{
		// printf("Removing factory for class %s, id %s, index %u\n", NameVector->elementAt(i), id, i);
		
		FactoryVector->elementAt(i)->DeleteFactory();
		FactoryVector->setElementAt((Mp7JrsFactory *)0, i);
		// We only remove identical name keys, Note: must be done before removing the name
		try{if(NameHashTable->get(X(NameVector->elementAt(i))))NameHashTable->removeKey(X(NameVector->elementAt(i)));}catch(...){}
		delete [] NameVector->elementAt(i); // We use delete[] instead of delete()
		NameVector->setElementAt((char *) 0, i);
		try{if(GuidHashTable->get(id))GuidHashTable->removeKey(id);}catch(...){}
		
	}
	FactoryVector->setElementAt(this, i);
	XMLCh * key = XERCES_CPP_NAMESPACE_QUALIFIER XMLString::transcode(classname); // Xerces valuehashtable cannot deal with non const keys
	NameHashTable->put(key, i);
	int len = (int)strlen(classname) + 1;
	char * name = new char[len]; // Vector uses delete(), so we must use new() not malloc()
	memcpy(name, classname, len);
	NameVector->setElementAt(name, i); 
	key = XERCES_CPP_NAMESPACE_QUALIFIER XMLString::transcode(id); // Xerces valuehashtable cannot deal with non const keys
	GuidHashTable->put(key, i);
	// printf("Adding factory for class %s, id %s, index %u\n", classname, id, i);
		
}

// Don't use delete, use DeleteFactory() instead;
Mp7JrsFactory::~Mp7JrsFactory(void)
{
}

void Mp7JrsFactory::Initialize()
{
	if(++initialized > 1)
		return;

	// Create static vectors and tables
	bool created = false;
	if(!FactoryVector)
	{
		FactoryVector = new Mp7JrsFactoryVector(128);
		created = true;
	}
	
	if(!GuidHashTable)
	{
		GuidHashTable = new Mp7JrsHashTable(3457); // Prime modulus
	}
	
	if(!NameHashTable)
	{
		NameHashTable = new Mp7JrsHashTable(3457); // Prime modulus
	}
	
	if(!NameVector)
	{
		// Since Xerces uses delete() instead of delete[](), 
		// it can't adopt the strings
		NameVector = new Mp7JrsNameVector(128, false); 
	}

	if(!SchemaLocationMap)
	{
		SchemaLocationMap = new Mp7JrsSchemaLocationMap();
	}
	
	if(created)
	{	
		// Insert the huge amount of core factories.
#		include "Mp7JrsFactoryInitialize.h"
	}
}


void Mp7JrsFactory::Terminate()
{
	if(--initialized > 0)
		return;

	unsigned int i;
	if(FactoryVector)
	{
		for(i = 0; i < FactoryVector->size(); ++i)
		{	
			if(FactoryVector->elementAt(i))
			{
			  // For unregistering extension libs in dlls, use the #include "myextlibDeinitialize.h"
				FactoryVector->elementAt(i)->DeleteFactory();
				FactoryVector->elementAt(i) = (Mp7JrsFactory *)0;
			}
		}
	}
	delete FactoryVector;
	FactoryVector = (Mp7JrsFactoryVector *)0;
		
	XMLCh* key;
	ValueHashTableOfEnumerator<unsigned int> e1(NameHashTable);
	while(e1.hasMoreElements())
	{
		key = (XMLCh*)e1.nextElementKey();
		//char * t = XMLString::transcode(key);
		//printf("Key = %s\n", t);
		//XMLString::release(&t);
		XMLString::release(&key);
	}
	delete NameHashTable;
	NameHashTable = (Mp7JrsHashTable *)0;
		
	ValueHashTableOfEnumerator<unsigned int> e2(GuidHashTable);
	while(e2.hasMoreElements())
	{
		key = (XMLCh*)e2.nextElementKey();
		//char * t = XMLString::transcode(key);
		//printf("Key = %s\n", t);
		//XMLString::release(&t);
		XMLString::release(&key);
	}
	delete GuidHashTable;
	GuidHashTable = (Mp7JrsHashTable *)0;

	for(i = 0; i < NameVector->size(); ++i)
	{	
		delete [] NameVector->elementAt(i); // We use delete[] instead of delete()
	}
	delete NameVector;
	NameVector = (Mp7JrsNameVector *)0;

	delete SchemaLocationMap;
	SchemaLocationMap = (Mp7JrsSchemaLocationMap *)0;
}

const char * Mp7JrsFactory::GetVersion()
{
	return "2.7.1";
}


const char * Mp7JrsFactory::GetModelVersion()
{
	return "2.1";
}

void Mp7JrsFactory::DeleteFactory()
{
	if(m_ToDelete)
	{
		delete this;
	}
}

Mp7JrsPtr<Mp7JrsNode> Mp7JrsFactory::CreateObject(unsigned int index)
{
	Mp7JrsPtr<Mp7JrsNode> result;
	if (!initialized) return result; // It's a null pointer at this point

	if(index > 0 && index < FactoryVector->size())
	{
		result = (FactoryVector->elementAt(index))->CreateObject();
		result->m_ClassId = index;

		// The node needs access to its own reference count sometimes -> set it.
		result->m_myself = result;
		result->Init();	// Call this after setting m_myself
	}
	else
	{
		char msg[512] = {0};
#if defined(XML_WIN32) || defined(WIN32) // Xerces 3.0 doesn't know XML_WIN32
		_snprintf_s(msg, 512, 511, "Mp7JrsFactory::CreateObject(): Sorry, no class registered at index %u.", index);
#elif defined(__APPLE__)// Xerces 3.0 doesn't know XML_APPLE
		sprintf(msg, "Mp7JrsFactory::CreateObject(): Sorry, no class registered at index %u.", index);
#elif defined(XML_LINUX) || defined(LINUX) || defined(linux) // Xerces 3.0 doesn't know XML_LINUX
		sprintf(msg, "Mp7JrsFactory::CreateObject(): Sorry, no class registered at index %u.", index);
#else // XML_LINUX ...
#	error "TODO: Add platform specific headers for xpathparsecontext."
#endif		
		throw Mp7JrsException(MP7JRS_ERROR, msg);
	}
	return result;
}

Mp7JrsNodePtr Mp7JrsFactory::CloneObject(const Mp7JrsNodePtr &original)
{
	if(original != NULL)
	{
		Mp7JrsNodePtr clone = Mp7JrsFactory::CreateObject(original->GetClassId());
		clone->DeepCopy(original);

		// Add Mp7JrsNode attribs
		if(original->GetContentName() != NULL)
		{
			clone->SetContentName(XMLString::replicate(original->GetContentName()));
		}
		clone->UseTypeAttribute = original->UseTypeAttribute;

		return clone;
	}
	else return Mp7JrsNodePtr();
}

unsigned int Mp7JrsFactory::GetTypeIndex(const XMLCh * qname)
{
	if(initialized)
	{
		unsigned int i = 1; // Unknown
		try
		{
			return NameHashTable->get(qname);
		}catch(...)
		{
			// Unknown	
		}
		return i;
	}
	throw Mp7JrsException(MP7JRS_ERROR,"Mp7JrsFactory is not initialised.");
}

const char * Mp7JrsFactory::GetTypeName(unsigned int index)
{
	if(initialized)
	{
		if(index < NameVector->size())
		{
			return NameVector->elementAt(index);
		}
		return GetTypeName(1); // Index 1 represents Unknown
	}
	throw Mp7JrsException(MP7JRS_ERROR,"Mp7JrsFactory is not initialised.");
}

void Mp7JrsFactory::DeleteObject(Mp7JrsNode * item)
{
	if(item)
	{
		(FactoryVector->elementAt(item->m_ClassId))->DeleteRawObject(item);
	}
}

void Mp7JrsFactory::DeleteRawObject(Mp7JrsNode * item)
{
	delete item;
}

//////////////////////////////////////////////////////////////////////
/*static*/ void Mp7JrsFactory::SetNamespaceUri(const XMLCh * prefix, const XMLCh * nsuri)
{
	SchemaLocationMap->SetSchemaLocation(nsuri, prefix, (XMLCh*)NULL);
}

/*static*/const XMLCh* Mp7JrsFactory::GetNamespaceUri(const XMLCh * prefix)
{
	const XMLCh* p = NULL;
	const XMLCh* ns = NULL;
	bool ok = SchemaLocationMap->GetFirstNamespace(&p, &ns);
	while (ok) {
		if(XMLString::compareString(p, prefix) == 0) {
			return ns;
		}
	if((p = SchemaLocationMap->GetNextSchemaPrefix(ns)) != 0) {
	  continue;
	}
		
		ok = SchemaLocationMap->GetNextNamespace(&p,&ns);
	}
	// Unknown types are mapped to default ns (which must be the first in list
	SchemaLocationMap->GetFirstNamespace(&p, &ns);
	return ns;
}

/*static*/ void Mp7JrsFactory::UnregisterFactory(SimpleGuid id)
{
	if(!FactoryVector) return;

	unsigned int i = MapGuidToIndex(id);
	if(i > 0 && i < FactoryVector->size())
	{
		if(FactoryVector->elementAt(i) && FactoryVector->elementAt(i)->GetToDelete())
		{
			FactoryVector->elementAt(i)->DeleteFactory();
			FactoryVector->elementAt(i) = (Mp7JrsFactory *)0;
		}
	}
}

