
// Based on XPathParseContext_cpp.template

/*****************************************************************************/
/* Copyright (C) 2003-2013, JOANNEUM RESEARCH Forschungs Ges.m.b.H.          */
/* Steyrergasse 17, A-8010 Graz, Austria.                                    */
/*                                                                           */
/*  This file is part of the JRS MPEG-7 Library.                             */
/*  http://mpeg-7.joanneum.at                                                */
/*                                                                           */
/*  JRS MPEG-7 Library is free software: you can redistribute it and/or      */
/*  modify it under the terms of the GNU Lesser General Public License       */
/*  (LGPL) as published by the Free Software Foundation, either version 3    */
/*  of the License, or (at your option) any later version.                   */
/*                                                                           */
/*  JRS MPEG-7 Library is distributed in the hope that it will be useful,    */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the             */
/*  GNU Lesser General Public License for more details.                      */
/*                                                                           */
/*  You should have received a copy of the GNU Lesser General Public License */
/*  along with this library. If not, see <http://www.gnu.org/licenses/>.     */
/*****************************************************************************/


#include <xercesc/util/regx/RegularExpression.hpp>
#include <xercesc/util/regx/Match.hpp>
#include <xercesc/dom/DOM.hpp>

#include "Mp7JrsUtil.h"
#include "Mp7JrsFactory.h"
#include "Mp7JrsMissingElementException.h" // Extended exception
#include "Mp7JrsXPathParseContext.h"
#include "Mp7JrsNodeCollection.h" // FTT Try to get rid of this

#include <stdio.h>
#include <assert.h>

// Intro of API documentation
#include "docintro.h"

Mp7JrsXPathParseContext::Mp7JrsXPathParseContext()
	: ElementName(NULL), XsiType(NULL), FreeIndexList(1), QualifiedXsiType(NULL)
{

}

/* virtual */ Mp7JrsXPathParseContext::~Mp7JrsXPathParseContext()
{
	XMLString::release(&ElementName); // reset to NULL is done in release()
	XMLString::release(&XsiType); // reset to NULL is done in release()
	XMLString::release(&QualifiedXsiType); // reset to NULL is done in release()
	XPath = NULL; // Not our buffer!
	UnparsedXPath = NULL; // Not our buffer!
}

/* static */ bool Mp7JrsXPathParseContext::Initialise(Mp7JrsXPathParseContext & c
	, Mp7JrsXPathParseContext ** context
	, const XMLCh *xpath
	, Mp7JrsXPathCreation flag)
{
	bool isentrypoint = false;
	if(!*context) // We are starting here
	{
		isentrypoint = true;
		
		c.Operation = MP7JRS_XPATH_UNKNOWN;
		c.Found = false;
		c.ElementName = NULL;
		c.XsiType = NULL;
		c.QualifiedXsiType = NULL;
		c.XPath = xpath;
		c.UnparsedXPath = NULL;
		c.Create = (flag == MP7JRS_FORCECREATE ? true : false);
		c.Id = 0; // 0 = Use default type
		c.CollectionSize = 0;
		c.UpperBoundary = -1; // -1 = Unbounded or not used
		
		c.PreferType = false;
		c.IsAttribute = false;
		c.StartIndex = -3; // Take everything
		c.StopIndex = -3; // Take everything
	
		c.ItemCount = 0; 
		c.ValidItemCount = 0;
		c.LastItemIndex = -1; // -1 = Unknown
		c.LastValidItemIndex = -1; // -1 = Unknown
		c.LastNode = Mp7JrsNodePtr();
		// c.FreeIndexList initialised automatically;
			
		*context = &c;
	}
	return isentrypoint;	
}


/* static */ void Mp7JrsXPathParseContext::Reset(Mp7JrsXPathParseContext * context)
{
	assert(context);
	context->Operation = MP7JRS_XPATH_UNKNOWN;
	XMLString::release(&context->ElementName); // reset to NULL is done in release()
	XMLString::release(&context->XsiType); // reset to NULL is done in release()
	XMLString::release(&context->QualifiedXsiType); // reset to NULL is done in release()
	// Dont touch XPath (global)
	context->UnparsedXPath = NULL;
	// Dont touch Create (global)
	// Dont touch IsAttribute
	// Dont touch Id;
	// Dont touch CollectionSize;
	// Dont touch UpperBoundary;
	
	context->PreferType = false;
	context->StartIndex = -3; // Take everything
	context->StopIndex = -3; // Take everything
	// Dont touch ItemCount;
	// Dont touch ValidItemCount;
	// Dont touch LastItemIndex;
	// Dont touch LastValidItemIndex;
	// Dont touch LastNode;
	context->FreeIndexList.removeAllElements();
}

/* static */ void Mp7JrsXPathParseContext::Parse(Mp7JrsXPathParseContext * context)
{
	if(!context) return;
	Mp7JrsXPathParseContext::Reset(context);
	
	if(!context->XPath) return;
	if(XMLString::stringLen(context->XPath) == 0) return;

#if 0
// XMLString::subString() test:
// Contrary to documentation subString end index is not inclusive in 2.7.0 and 2.2.0 !!!
	XMLCh * t1 = XMLString::transcode("01234");
	XMLCh * t2 = XMLString::transcode("	 ");
	XMLString::subString(t2, t1, 0, 3); // According to docu should reveal "0123"
	if(! XMLString::endsWith(t2, X("3")))
	{
		bool bloody_docu = true;
	}
#endif	
	
	context->UnparsedXPath = const_cast<XMLCh *> (context->XPath);
	int state = 1;
	
	while(state > 0)
	{
		switch(state)
		{
			case 1:	state = Mp7JrsXPathParseContext::ScanSlash(context) ? 0 : 2; break;
			case 2: state = Mp7JrsXPathParseContext::ScanUp(context) ? 0 : 3; break;
			case 3: Mp7JrsXPathParseContext::ScanAttribute(context); state = 4; break;
			case 4: state = Mp7JrsXPathParseContext::ScanElement(context) ? 5 : -1; break;
			case 5: state = Mp7JrsXPathParseContext::ScanTail(context) ? 0: 6; break;
			case 6: state = Mp7JrsXPathParseContext::ScanIndex(context) ? 7 : 8; break;
			case 7: state = Mp7JrsXPathParseContext::ScanXsiType(context) ? 11 : 11; break;
			case 8: state = Mp7JrsXPathParseContext::ScanXsiType(context) ? 9 : 10; break;
			case 9: state = Mp7JrsXPathParseContext::ScanIndex(context) ? 10 : 10; break;
			
			case 10: 
				context->PreferType = true;
				state = 11;
				 break;
				 
			case 11:
			 	if(!Mp7JrsXPathParseContext::ScanTail(context))
			 	{
					XMLCh * msg = XMLString::transcode("XPath error:\n");
					Mp7JrsUtil::CatString(&msg, context->XPath);
					Mp7JrsUtil::CatString(&msg, X("\nSorry, unexpected expression \""));
					Mp7JrsUtil::CatString(&msg, context->UnparsedXPath);
					Mp7JrsUtil::CatString(&msg, X("\"."));
					Mp7JrsException e(MP7JRS_ERROR, msg);
					XMLString::release(&msg);
					throw e;
				}
				state = 0;
				break;

			default: break; // Exit
		}
	}
}

/* static */ bool Mp7JrsXPathParseContext::ScanSlash(Mp7JrsXPathParseContext * context)
{
	assert(context);
	if(XMLString::startsWith(context->UnparsedXPath, X("/")))
	{
		if(XMLString::startsWith(context->UnparsedXPath, X("//")))
		{
			XMLCh * msg = XMLString::transcode("XPath error:\n");
			Mp7JrsUtil::CatString(&msg, context->XPath);
			Mp7JrsUtil::CatString(&msg, X("\nSorry, \"//\" is not supported."));
			Mp7JrsException e(MP7JRS_ERROR, msg);
			XMLString::release(&msg);
			throw e;
		}
		context->Operation = MP7JRS_XPATH_ROOT;
		++context->UnparsedXPath; // Eat up /
		while(XMLString::startsWith(context->UnparsedXPath, X(" "))) // And blanks
			++context->UnparsedXPath;
		
		return true;
	}
	return false;
}

/* static */ bool Mp7JrsXPathParseContext::ScanUp(Mp7JrsXPathParseContext * context)
{
	assert(context);
	if(XMLString::startsWith(context->UnparsedXPath, X("..")))
	{
		context->UnparsedXPath += 2;
		if(XMLString::startsWith(context->UnparsedXPath, X("/")))
		{
			if(XMLString::startsWith(context->UnparsedXPath, X("//")))
			{
				XMLCh * msg = XMLString::transcode("XPath error:\n");
				Mp7JrsUtil::CatString(&msg, context->XPath);
				Mp7JrsUtil::CatString(&msg, X("\nSorry, \"//\" is not supported."));
				Mp7JrsException e(MP7JRS_ERROR, msg);
				XMLString::release(&msg);
				throw e;
			}
			++context->UnparsedXPath; // Eat up /
		}
		while(XMLString::startsWith(context->UnparsedXPath, X(" "))) // And blanks
			++context->UnparsedXPath;
		context->Operation = MP7JRS_XPATH_UP;
		return true;
	}
	return false;
}

/* static */ bool Mp7JrsXPathParseContext::ScanAttribute(Mp7JrsXPathParseContext * context)
{
	assert(context);
	
	Match m;
	RegularExpression r("^@"); // Matches '@' to indicate an attribute
	if(r.matches(context->UnparsedXPath, &m))
	{
		context->IsAttribute = true;
		context->Operation = MP7JRS_XPATH_SELECT_ALL;
		context->UnparsedXPath += 1;
		return true;
	}
	return false;
}

/* static */ bool Mp7JrsXPathParseContext::ScanElement(Mp7JrsXPathParseContext * context)
{
	assert(context);
	
	Match m;
	RegularExpression r("^([0-9a-zA-Z_:]+)"); // Matches 'bla', where bla can be prefixed
	if(r.matches(context->UnparsedXPath, &m))
	{
		context->Operation = MP7JRS_XPATH_SELECT_ALL;
		int len = m.getEndPos(1) - m.getStartPos(1) + 1;
		context->ElementName = Mp7JrsUtil::allocate(len);
		XMLString::subString(context->ElementName, context->UnparsedXPath, m.getStartPos(1), m.getEndPos(1));
		// Strip prefix if exists, TODO: this is hacky
		XMLCh * tmp = XMLString::replicate(Mp7JrsUtil::LocalName(context->ElementName));
		Mp7JrsUtil::deallocate(&context->ElementName);
		context->ElementName = tmp;
		context->UnparsedXPath += m.getEndPos(0);
	}
	else
	{
		XMLCh * msg = XMLString::transcode("XPath error:\n");
		Mp7JrsUtil::CatString(&msg, context->XPath);
		Mp7JrsUtil::CatString(&msg, X("\nValid element name [0-9a-zA-Z_:]+ expected."));
		Mp7JrsException e(MP7JRS_ERROR, msg);
		XMLString::release(&msg);
		throw e;			
	}
	return true;
}

/* static */ bool Mp7JrsXPathParseContext::ScanTail(Mp7JrsXPathParseContext * context)
{
	assert(context);

	if(XMLString::stringLen(context->UnparsedXPath) > 0)
	{
		if(context->IsAttribute)
		{
			XMLCh * msg = XMLString::transcode("XPath error:\n");
			Mp7JrsUtil::CatString(&msg, context->XPath);
			Mp7JrsUtil::CatString(&msg, X("\nSorry, only plain attributes (e.g. @foo) are supported."));
			Mp7JrsException e(MP7JRS_ERROR, msg);
			XMLString::release(&msg);
			throw e;
		}
		if(XMLString::startsWith(context->UnparsedXPath, X("/")))
		{
			if(XMLString::startsWith(context->UnparsedXPath, X("//")))
			{
				XMLCh * msg = XMLString::transcode("XPath error:\n");
				Mp7JrsUtil::CatString(&msg, context->XPath);
				Mp7JrsUtil::CatString(&msg, X("\nSorry, \"//\" is not supported."));
				Mp7JrsException e(MP7JRS_ERROR, msg);
				XMLString::release(&msg);
				throw e;			
			}
			if(XMLString::endsWith(context->UnparsedXPath, X("/")))
			{
				XMLCh * msg = XMLString::transcode("XPath error:\n");
				Mp7JrsUtil::CatString(&msg, context->XPath);
				Mp7JrsUtil::CatString(&msg, X("\nSorry, unexpected end."));
				Mp7JrsException e(MP7JRS_ERROR, msg);
				XMLString::release(&msg);
				throw e;			
			}
			++context->UnparsedXPath; // Consume slash
		}
		else
		{
			return false; // Maybe an index or xsi:type
		}
	}
	return true;	
}

/* static */ bool Mp7JrsXPathParseContext::ScanIndex(Mp7JrsXPathParseContext * context)
{
	assert(context);
	
// This is odd! Xerces 2.7. can deal with this regular expression, whilst 2.2 can manage just half of the cases
// I don't have time to dig into this, so here are two solutions,
// the smart one for 2.7 and the dumb for 2.2!
#if _XERCES_VERSION >= 20700

  	Match m;
  	// Matches: [123] or [last()] or [position()>1 | position()<123] 
  	// or [position()<123] or [position()>1] or [position()>last()]
	// Note, order is important. The long position() expression must come first!
	RegularExpression r(
		"^[\t ]*\\[[\t ]*"
		"("
		"([0-9]+)"
		"|(last[\t ]*\\([\t ]*\\))"
		"|(position[\t ]*\\([\t ]*\\)[\t ]*>[\t ]*([0-9]+)[\t ]*\\|[\t ]*position\\(\\)[\t ]*<[\t ]*([0-9]+))" // Must come first
		"|(position[\t ]*\\([\t ]*\\)[\t ]*<[\t ]*([0-9]+))"
		"|(position[\t ]*\\([\t ]*\\)[\t ]*>[\t ]*([0-9]+))"
		"|(position[\t ]*\\([\t ]*\\)[\t ]*>[\t ]*last[\t ]*\\([\t ]*\\))"
		")"
		"[\t ]*\\][\t ]*"); 
	
	if(r.matches(context->UnparsedXPath, &m))
	{
		// Indexes for getStartPos():
		// Group index 0 is for the whole expression
		// Group index 1 is for the first (..) term (i.e. (
		// Group index 2 is for the second (..) term  (i.e ([0-9]+) ...
	
		context->Operation = MP7JRS_XPATH_SELECT_ALL;
		if(m.getEndPos(2) - m.getStartPos(2) > 0) // [123] group
		{
			int n;
			if(Mp7JrsXPathParseContext::ConvertIndex(m.getStartPos(2), m.getEndPos(2), 0, context, n))
			{
				context->Operation = context->XsiType ? MP7JRS_XPATH_SELECT_TYPED_INDEX : MP7JRS_XPATH_SELECT_INDEX;
				context->StartIndex = context->StopIndex = n;
				context->UnparsedXPath += m.getEndPos(0);
				return true;
			}
			return false;
		}
		else if(m.getEndPos(3) - m.getStartPos(3) > 0) // [last()] group
		{
			context->Operation = context->XsiType ? MP7JRS_XPATH_SELECT_TYPED_LAST : MP7JRS_XPATH_SELECT_LAST;
			context->StartIndex = context->StopIndex = -2; // -2 Indicates 'last()'
			context->UnparsedXPath += m.getEndPos(0);
			return true;
		}
		else if(m.getEndPos(4) - m.getStartPos(4) > 0) // [position()>123 | position()<123] group
		{
			if(context->Create)
			{
				Mp7JrsXPathParseContext::ErrorUnallowedExpression(context);
			}
			int n;
			if(!Mp7JrsXPathParseContext::ConvertIndex(m.getStartPos(5), m.getEndPos(5), -1, context, n))
			{
				return false;
			}
			context->StartIndex = n;
			
			if(!Mp7JrsXPathParseContext::ConvertIndex(m.getStartPos(6), m.getEndPos(6), -1, context, n))
			{
				return false;
			}
			context->Operation = context->XsiType ? MP7JRS_XPATH_SELECT_TYPED_POSITION_RANGE : MP7JRS_XPATH_SELECT_POSITION_RANGE;
			context->StopIndex = n;
			context->UnparsedXPath += m.getEndPos(0);
			return true;
		}
		else if(m.getEndPos(7) - m.getStartPos(7) > 0) // [position()<123] group
		{
			if(context->Create)
			{
				Mp7JrsXPathParseContext::ErrorUnallowedExpression(context);
			}
			int n;
			if(!Mp7JrsXPathParseContext::ConvertIndex(m.getStartPos(8), m.getEndPos(8), -1, context, n))
			{
				return false;
			}
			context->Operation = context->XsiType ? MP7JRS_XPATH_SELECT_TYPED_POSITION_LT : MP7JRS_XPATH_SELECT_POSITION_LT;
			context->StartIndex = -1;
			context->StopIndex = n; 
			context->UnparsedXPath += m.getEndPos(0);
			return true;
		}
		else if(m.getEndPos(9) - m.getStartPos(9) > 0) // [position()>123] group
		{
			if(context->Create)
			{
				Mp7JrsXPathParseContext::ErrorUnallowedExpression(context);
			}
			int n;
			if(!Mp7JrsXPathParseContext::ConvertIndex(m.getStartPos(10), m.getEndPos(10), -1, context, n))
			{
				return false;
			}
			context->Operation = context->XsiType ? MP7JRS_XPATH_SELECT_TYPED_POSITION_GT : MP7JRS_XPATH_SELECT_POSITION_GT;
			context->StartIndex = n;
			context->StopIndex = -2; // last()
			context->UnparsedXPath += m.getEndPos(0);
			return true;
		}
		else if(m.getEndPos(11) - m.getStartPos(11) > 0) // [position()>last()] group
		{
			if(context->Create)
			{
				context->Operation = MP7JRS_XPATH_FORCE_CREATE;
			}
			context->UnparsedXPath += m.getEndPos(0);
			return true;
		}
		else assert(false);
	}
#else // Xerces 2.2.0
  	Match m;
  	// Matches: [123] or [last()] or [position()>1 | position()<123] 
  	// or [position()<123] or [position()>1] or [position()>last()]
	// Note, order is important. The long position() expression must come first!
	RegularExpression r1(
		"^[\t ]*\\[[\t ]*"
		"([0-9]+)"
		"[\t ]*\\][\t ]*"); 
	if(r1.matches(context->UnparsedXPath, &m))
	{
		context->Operation = MP7JRS_XPATH_SELECT_ALL;
		if(m.getEndPos(1) - m.getStartPos(1) > 0) // [123] group
		{
			int n;
			if(Mp7JrsXPathParseContext::ConvertIndex(m.getStartPos(1), m.getEndPos(1), 0, context, n))
			{
				context->Operation = context->XsiType ? MP7JRS_XPATH_SELECT_TYPED_INDEX : MP7JRS_XPATH_SELECT_INDEX;
				context->StartIndex = context->StopIndex = n;
				context->UnparsedXPath += m.getEndPos(0);
				return true;
			}
			return false;
		}
	}
	RegularExpression r2(
		"^[\t ]*\\[[\t ]*"
		"last[\t ]*\\([\t ]*\\)"
		"[\t ]*\\][\t ]*"); 
	if(r2.matches(context->UnparsedXPath, &m)) // [last()]
	{
		context->Operation = context->XsiType ? MP7JRS_XPATH_SELECT_TYPED_LAST : MP7JRS_XPATH_SELECT_LAST;
		context->StartIndex = context->StopIndex = -2; // -2 Indicates 'last()'
		context->UnparsedXPath += m.getEndPos(0);
		return true;
	}
	RegularExpression r3(
		"^[\t ]*\\[[\t ]*"
		"(position[\t ]*\\([\t ]*\\)[\t ]*>[\t ]*([0-9]+)[\t ]*\\|[\t ]*position\\(\\)[\t ]*<[\t ]*([0-9]+))"
		"[\t ]*\\][\t ]*"); 
	if(r3.matches(context->UnparsedXPath, &m)) // [position()>123 | position()<123]
	{
		if(m.getEndPos(1) - m.getStartPos(1) > 0) // [position()>123 | position()<123] group
		{
			if(context->Create)
			{
				Mp7JrsXPathParseContext::ErrorUnallowedExpression(context);
			}
			int n;
			if(!Mp7JrsXPathParseContext::ConvertIndex(m.getStartPos(2), m.getEndPos(2), -1, context, n))
			{
				return false;
			}
			context->StartIndex = n;
			
			if(!Mp7JrsXPathParseContext::ConvertIndex(m.getStartPos(3), m.getEndPos(3), -1, context, n))
			{
				return false;
			}
			context->Operation = context->XsiType ? MP7JRS_XPATH_SELECT_TYPED_POSITION_RANGE : MP7JRS_XPATH_SELECT_POSITION_RANGE;
			context->StopIndex = n;
			context->UnparsedXPath += m.getEndPos(0);
			return true;
		}
	}
	RegularExpression r4(
		"^[\t ]*\\[[\t ]*"
		"position[\t ]*\\([\t ]*\\)[\t ]*<[\t ]*([0-9]+)"
		"[\t ]*\\][\t ]*"); 
	if(r4.matches(context->UnparsedXPath, &m)) // [position()<123]
	{
		if(context->Create)
		{
			Mp7JrsXPathParseContext::ErrorUnallowedExpression(context);
		}
		int n;
		if(!Mp7JrsXPathParseContext::ConvertIndex(m.getStartPos(1), m.getEndPos(1), -1, context, n))
		{
			return false;
		}
		context->Operation = context->XsiType ? MP7JRS_XPATH_SELECT_TYPED_POSITION_LT : MP7JRS_XPATH_SELECT_POSITION_LT;
		context->StartIndex = -1;
		context->StopIndex = n; 
		context->UnparsedXPath += m.getEndPos(0);
		return true;
	}
	RegularExpression r5(
		"^[\t ]*\\[[\t ]*"
		"position[\t ]*\\([\t ]*\\)[\t ]*>[\t ]*([0-9]+)"
		"[\t ]*\\][\t ]*"); 
	if(r5.matches(context->UnparsedXPath, &m)) // [position()>123] 
	{
		if(context->Create)
		{
			Mp7JrsXPathParseContext::ErrorUnallowedExpression(context);
		}
		int n;
		if(!Mp7JrsXPathParseContext::ConvertIndex(m.getStartPos(1), m.getEndPos(1), -1, context, n))
		{
			return false;
		}
		context->Operation = context->XsiType ? MP7JRS_XPATH_SELECT_TYPED_POSITION_GT : MP7JRS_XPATH_SELECT_POSITION_GT;
		context->StartIndex = n;
		context->StopIndex = -2; // last()
		context->UnparsedXPath += m.getEndPos(0);
		return true;
	}
	RegularExpression r6(
		"^[\t ]*\\[[\t ]*"
		"position[\t ]*\\([\t ]*\\)[\t ]*>[\t ]*last[\t ]*\\([\t ]*\\)"
		"[\t ]*\\][\t ]*"); 
	if(r6.matches(context->UnparsedXPath, &m)) // [position()>last()]
	{
		if(context->Create)
		{
			context->Operation = MP7JRS_XPATH_FORCE_CREATE;
		}
		context->UnparsedXPath += m.getEndPos(0);
		return true;
	}
#endif // 2.7 vs 2.2

   	return false;
}

/* static */ bool Mp7JrsXPathParseContext::ScanXsiType(Mp7JrsXPathParseContext * context)
{
  assert(context);
	
  Match m; // Matches: [@xsi:type = 'bla'] where 'bla' can be prefixed
	RegularExpression r(
		"^[\t ]*\\[[\t ]*"
//		"\\@xsi\\:type[\t ]*=[\t ]*[\\'\\\"]([0-9a-zA-Z_]+)[\\'\\\"]" // Works in 2.7.0, but not in 3.0.1 anymore
		"@xsi:type[\t ]*=[\t ]*[\'\"]([0-9a-zA-Z_:]+)[\'\"]"
		"[\t ]*\\][\t ]*");

  if(r.matches(context->UnparsedXPath, &m))
	{
		context->Operation = (
			context->Operation == MP7JRS_XPATH_SELECT_ALL ? MP7JRS_XPATH_SELECT_TYPED_ALL
			: context->Operation == MP7JRS_XPATH_SELECT_INDEX ? MP7JRS_XPATH_SELECT_INDEXED_TYPE
			: context->Operation == MP7JRS_XPATH_SELECT_LAST ? MP7JRS_XPATH_SELECT_LAST_TYPE
			: context->Operation == MP7JRS_XPATH_SELECT_POSITION_RANGE ? MP7JRS_XPATH_SELECT_POSITION_RANGE_TYPE
			: context->Operation == MP7JRS_XPATH_SELECT_POSITION_LT ? MP7JRS_XPATH_SELECT_POSITION_LT_TYPE
			: context->Operation == MP7JRS_XPATH_SELECT_POSITION_GT ? MP7JRS_XPATH_SELECT_POSITION_GT_TYPE
			: context->Operation == MP7JRS_XPATH_FORCE_CREATE ? MP7JRS_XPATH_FORCE_CREATE
			: MP7JRS_XPATH_UNKNOWN);
		
		int len = m.getEndPos(1) - m.getStartPos(1) + 1;
		context->XsiType = Mp7JrsUtil::allocate(len);
		XMLString::subString(context->XsiType, context->UnparsedXPath, m.getStartPos(1), m.getEndPos(1));
		if(context->XsiType != NULL && XMLString::stringLen(context->XsiType) > 0){
		  if(context->QualifiedXsiType){
			XMLString::release(&context->QualifiedXsiType);
		  }
		  context->QualifiedXsiType = Mp7JrsXPathParseContext::ResolveTypeName(context->XsiType);
		}
		context->UnparsedXPath += m.getEndPos(0);
		return true;
	}
   	return false;
}

/* static */ bool Mp7JrsXPathParseContext::ConvertIndex(int startpos, int endpos, int minval, Mp7JrsXPathParseContext * context, int & n)
{
	assert(context);
	
	int len = endpos - startpos + 1;
	XMLCh * tmp = Mp7JrsUtil::allocate(len);
	XMLString::subString(tmp, context->UnparsedXPath, startpos, endpos);
	n = XMLString::parseInt(tmp) - 1; // Convert to 0 based C++ index
	if(n < minval)
	{
		char buf[32] = {0};
		XMLCh * msg = XMLString::transcode("XPath error:\n");
		Mp7JrsUtil::CatString(&msg, context->XPath);
		Mp7JrsUtil::CatString(&msg, X("\nSorry, invalid index "));
		Mp7JrsUtil::CatString(&msg, tmp);
		Mp7JrsUtil::CatString(&msg, X("\nValid minimum value: "));
#if defined(XML_WIN32) || defined(WIN32) // Xerces 3.0 doesn't know XML_WIN32
		_snprintf_s(buf, 32, 31, "%d", minval + 1); // Back to (1 based) XPath notation
#elif defined(__APPLE__) || defined(APPLE) || defined(_APPLE) // Xerces 3.0 doesn't know XML_LINUX
		sprintf(buf, "%d", minval + 1); // Back to (1 based) XPath notation
#elif defined(XML_LINUX) || defined(LINUX) || defined(linux) // Xerces 3.0 doesn't know XML_LINUX
		sprintf(buf, "%d", minval + 1); // Back to (1 based) XPath notation
#else // XML_LINUX ...
#	error "TODO: Add platform specific headers for xpathparsecontext."
#endif
		Mp7JrsUtil::CatString(&msg, X(buf));
		Mp7JrsException e(MP7JRS_ERROR, msg);
		XMLString::release(&msg);
		XMLString::release(&tmp);
		throw e;			
	}
	XMLString::release(&tmp);
	return true;
}


/* static */ Mp7JrsNodePtr Mp7JrsXPathParseContext::CreateNode(Mp7JrsXPathParseContext * context, unsigned int defaulttype)
{
	assert(context);
	
	assert(defaulttype != 1); // Must not be Unkown
	unsigned int id = defaulttype;
	if(context->XsiType)
	{
		id = Mp7JrsFactory::GetTypeIndex(context->QualifiedXsiType);
		if(id == 1) // Oh oh, unknown
		{
			// Is it our Unknown type then ok, we create an Unknown node
			// Otherwise it might be a typo and we issue an error
			XMLCh * unknown = XMLString::transcode(Mp7JrsFactory::GetTypeName(1)); // Our Unknown is always at index 1
			if(XMLString::compareString(context->XsiType, unknown) != 0)
			{
				XMLString::release(&unknown);
				XMLCh * msg = XMLString::transcode("XPath error:\n");
				Mp7JrsUtil::CatString(&msg, context->XPath);
				Mp7JrsUtil::CatString(&msg, X("\nSorry, unknown xsi:type \""));
				Mp7JrsUtil::CatString(&msg, context->XsiType);
				Mp7JrsUtil::CatString(&msg, X("\"\nCannot create this type."));
				Mp7JrsException e(MP7JRS_ERROR, msg);
				XMLString::release(&msg);
				throw e;
			}
			XMLString::release(&unknown);
		}
	}
	else if(id < 1) // We have an abstract default type
	{
		Mp7JrsXPathParseContext::ErrorAbstractType(context);
	}
	
	Mp7JrsNodePtr node = Mp7JrsFactory::CreateObject(id);	
	if(node == Mp7JrsNodePtr())
	{
		char buf[32] = {0};
#if defined(XML_WIN32) || defined(WIN32) // Xerces 3.0 doesn't know XML_WIN32
		_snprintf_s(buf, 32, 31, "%u", defaulttype);
#elif defined(__APPLE__) || defined(APPLE) || defined(_APPLE) // Xerces 3.0 doesn't know XML_LINUX
		sprintf(buf, "%u", defaulttype);
#elif defined(XML_LINUX) || defined(LINUX) || defined(linux) // Xerces 3.0 doesn't know XML_LINUX
		sprintf(buf, "%u", defaulttype);
#else // XML_LINUX ...
#	error "TODO: Add platform specific headers for xpathparsecontext."
#endif		
		XMLCh * msg = XMLString::transcode("XPath error:\n");
		Mp7JrsUtil::CatString(&msg, context->XPath);
		Mp7JrsUtil::CatString(&msg, X("\nSorry, unknown default type with id "));
		Mp7JrsUtil::CatString(&msg, X(buf));
		Mp7JrsUtil::CatString(&msg, X(".\nCannot create this node."));
		Mp7JrsException e(MP7JRS_ERROR, msg);
		XMLString::release(&msg);
		throw e;
	}
	if(context->XsiType)
	{
		node->UseTypeAttribute = true;
	}
	return node;
}

/* static */ void Mp7JrsXPathParseContext::PreProcessSimpleAttribute(Mp7JrsXPathParseContext * context)
{
	Mp7JrsXPathParseContext::PreProcessSimpleElement(context); // All handled in PreProcessSimpleElement();
}

/* static */ void Mp7JrsXPathParseContext::PreProcessSimpleElement(Mp7JrsXPathParseContext * context)
{
	assert(context);
	
	switch(context->Operation)
	{
		case MP7JRS_XPATH_SELECT_ALL: // 
		case MP7JRS_XPATH_SELECT_TYPED_ALL: // [@xsi:type='bla']
			break; // Ok
			
		case MP7JRS_XPATH_SELECT_INDEX: // [i]
		case MP7JRS_XPATH_SELECT_INDEXED_TYPE: // [i @xsi:type]
		case MP7JRS_XPATH_SELECT_TYPED_INDEX: // [@xsi:type i]
			if(context->Create && context->StartIndex != 0)
			{
				Mp7JrsXPathParseContext::ErrorInvalidIndex(context);
			}
			break;
		
		case MP7JRS_XPATH_SELECT_LAST: // [last()]
		case MP7JRS_XPATH_SELECT_LAST_TYPE: // [last() xsi:type]
		case MP7JRS_XPATH_SELECT_TYPED_LAST: // [xsi:type last()]
			break; // ok
					
		case MP7JRS_XPATH_SELECT_POSITION_RANGE: // [position()>A | position()<B]
		case MP7JRS_XPATH_SELECT_POSITION_RANGE_TYPE: // [position()>A | position()<B xsi:type]
		case MP7JRS_XPATH_SELECT_TYPED_POSITION_RANGE: // [xsi:type position()>A | position()<B]
			if(context->Create && (context->StartIndex > -1 || context->StopIndex < 1))
			{
				Mp7JrsXPathParseContext::ErrorInvalidIndex(context);
			}
			break;
		
		case MP7JRS_XPATH_SELECT_POSITION_LT: // [position()<A]
		case MP7JRS_XPATH_SELECT_POSITION_LT_TYPE: // [position()<A xsi:type]
		case MP7JRS_XPATH_SELECT_TYPED_POSITION_LT: // [xsi:type position()<A]
			if(context->Create && context->StopIndex < 1)
			{
				Mp7JrsXPathParseContext::ErrorInvalidIndex(context);
			}
			break;
		
		case MP7JRS_XPATH_SELECT_POSITION_GT: // [position()>A]
		case MP7JRS_XPATH_SELECT_POSITION_GT_TYPE: // [position()>A xsi:type]
		case MP7JRS_XPATH_SELECT_TYPED_POSITION_GT: // [xsi:type position()>A]
			if(context->Create && context->StartIndex > 0)
			{
				Mp7JrsXPathParseContext::ErrorInvalidIndex(context);
			}
			break;

		//case MP7JRS_XPATH_FORCE_CREATE:
		
		default:;
	}	
}

/* static */ bool Mp7JrsXPathParseContext::HandleSimpleAttribute(Mp7JrsXPathParseContext * context, Mp7JrsNodePtr element)
{
	return Mp7JrsXPathParseContext::HandleSimpleElement(context, element); // All handled in HandleSimpleElement();
}

/* static */ bool Mp7JrsXPathParseContext::HandleSimpleElement(Mp7JrsXPathParseContext * context, Mp7JrsNodePtr element)
{
	assert(context);
	
	switch(context->Operation)
	{
		case MP7JRS_XPATH_SELECT_ALL: //
		case MP7JRS_XPATH_SELECT_INDEX: // [i]
		case MP7JRS_XPATH_SELECT_LAST: // [last()]
			return true;
			
		case MP7JRS_XPATH_SELECT_TYPED_ALL: // [@xsi:type='bla']
		case MP7JRS_XPATH_SELECT_INDEXED_TYPE: // [i @xsi:type]
		case MP7JRS_XPATH_SELECT_TYPED_INDEX: // [@xsi:type i]
		case MP7JRS_XPATH_SELECT_LAST_TYPE: // [last() xsi:type]
		case MP7JRS_XPATH_SELECT_TYPED_LAST: // [xsi:type last()]
			return(element->GetClassId() == Mp7JrsFactory::GetTypeIndex(context->QualifiedXsiType));
					
		case MP7JRS_XPATH_SELECT_POSITION_RANGE: // [position()>A | position()<B]
		case MP7JRS_XPATH_SELECT_POSITION_RANGE_TYPE: // [position()>A | position()<B xsi:type]
		case MP7JRS_XPATH_SELECT_TYPED_POSITION_RANGE: // [xsi:type position()>A | position()<B]
		case MP7JRS_XPATH_SELECT_POSITION_LT: // [position()<A]
		case MP7JRS_XPATH_SELECT_POSITION_LT_TYPE: // [position()<A xsi:type]
		case MP7JRS_XPATH_SELECT_TYPED_POSITION_LT: // [xsi:type position()<A]
		case MP7JRS_XPATH_SELECT_POSITION_GT: // [position()>A]
		case MP7JRS_XPATH_SELECT_POSITION_GT_TYPE: // [position()>A xsi:type]
		case MP7JRS_XPATH_SELECT_TYPED_POSITION_GT: // [xsi:type position()>A]
			return true; // Already checked in PreProcessSimpleElement

		case MP7JRS_XPATH_FORCE_CREATE: // [position() > last()]
			if(context->Create) // You cant beat last()
			{
				context->UpperBoundary = 1;
				Mp7JrsXPathParseContext::ErrorBoundary(context);
			}
			return false;
			break;
		default:;
	}
	return true;
}

/* static */ bool Mp7JrsXPathParseContext::PreProcessCollection(Mp7JrsXPathParseContext * context, unsigned int collsize, int upperboundary)
{
	assert(context);
	
	context->CollectionSize = collsize;
	context->Id = context->XsiType ? Mp7JrsFactory::GetTypeIndex(context->QualifiedXsiType) : 0;
	context->UpperBoundary = upperboundary;
	
	switch(context->Operation)
	{
		case MP7JRS_XPATH_SELECT_POSITION_LT: // [position()<A]
		case MP7JRS_XPATH_SELECT_POSITION_LT_TYPE: // [position()<A xsi:type]
		case MP7JRS_XPATH_SELECT_TYPED_POSITION_LT: // [xsi:type position()<A]
			context->StartIndex = -1; 
			break;
		case MP7JRS_XPATH_SELECT_POSITION_GT: // [position()>B]
		case MP7JRS_XPATH_SELECT_POSITION_GT_TYPE: // [position()>B xsi:type]
		case MP7JRS_XPATH_SELECT_TYPED_POSITION_GT: // [xsi:type position()>B]
			context->StopIndex = collsize; 
			break;
		case MP7JRS_XPATH_SELECT_POSITION_RANGE: // [position()>A | position()<B]
		case MP7JRS_XPATH_SELECT_POSITION_RANGE_TYPE: // [position()>A | position()<B xsi:type]
		case MP7JRS_XPATH_SELECT_TYPED_POSITION_RANGE: // [xsi:type position()>A | position()<B]
			if(context->StartIndex >= context->StopIndex) return false; // Logical exclusion
			break;
		default:;
	}
	return true;
}
/* static */ int Mp7JrsXPathParseContext::HandleCollection(Mp7JrsXPathParseContext * context, int i, Mp7JrsNodePtr element, Mp7JrsNodeEnum & result)
{
	assert(context);
	
	if(context->Operation == MP7JRS_XPATH_FORCE_CREATE)
	{
		return context->CollectionSize; // Finished
	}

	++context->ItemCount;
	context->LastItemIndex = i;
	
	switch(context->Operation)
	{
		case MP7JRS_XPATH_SELECT_ALL: // 
		case MP7JRS_XPATH_SELECT_TYPED_ALL: // [@xsi:type='bla']
			if(!context->Id || context->Id == element->GetClassId())
			{
				result.Insert(element);
				context->LastValidItemIndex = i;
				++context->ValidItemCount;
				return ++i;
			}
			break;
		
		case MP7JRS_XPATH_SELECT_INDEX: // [i]
		case MP7JRS_XPATH_SELECT_INDEXED_TYPE: // [i @xsi:type]
			if(context->ItemCount == context->StartIndex + 1)
			{
				if(!context->Id || context->Id == element->GetClassId())
				{
					result.Insert(element);
					context->LastValidItemIndex = i;
					++context->ValidItemCount;
				}
				return context->CollectionSize; // Finished
			}
			break;
		
		case MP7JRS_XPATH_SELECT_TYPED_INDEX: // [@xsi:type i]
			if(!context->Id || context->Id == element->GetClassId())
			{
				context->LastValidItemIndex = i;
				++context->ValidItemCount;
				if(context->ValidItemCount == context->StartIndex + 1)
				{
					result.Insert(element);
					return context->CollectionSize; // Finished
				}
			}
			break;
		
		case MP7JRS_XPATH_SELECT_LAST: // [last()]
		case MP7JRS_XPATH_SELECT_LAST_TYPE: // [last() xsi:type]
		case MP7JRS_XPATH_SELECT_TYPED_LAST: // [xsi:type last()]
			if(!context->Id || context->Id == element->GetClassId())
			{
				context->LastNode = element; // Evaluated at PostProcessCollection()
				context->LastValidItemIndex = i;
				++context->ValidItemCount;
			}
			break;
					
		case MP7JRS_XPATH_SELECT_POSITION_RANGE: // [position()>A | position()<B]
		case MP7JRS_XPATH_SELECT_POSITION_RANGE_TYPE: // [position()>A | position()<B xsi:type]
		case MP7JRS_XPATH_SELECT_TYPED_POSITION_RANGE: // [xsi:type position()>A | position()<B]
		case MP7JRS_XPATH_SELECT_POSITION_LT: // [position()<A]
		case MP7JRS_XPATH_SELECT_POSITION_LT_TYPE: // [position()<A xsi:type]
		case MP7JRS_XPATH_SELECT_TYPED_POSITION_LT: // [xsi:type position()<A]
		case MP7JRS_XPATH_SELECT_POSITION_GT: // [position()>A]
		case MP7JRS_XPATH_SELECT_POSITION_GT_TYPE: // [position()>A xsi:type]
		case MP7JRS_XPATH_SELECT_TYPED_POSITION_GT: // [xsi:type position()>A]
			if(i > context->StartIndex)
			{
				if(i < context->StopIndex)
				{
					if(!context->Id || context->Id == element->GetClassId())
					{
						result.Insert(element);
						context->LastValidItemIndex = i;
						++context->ValidItemCount;
					}
				} else return context->CollectionSize; // Finished
			}
			break;
			
		default:;	
	}		
	return ++i;
}

/* static */ bool Mp7JrsXPathParseContext::PostProcessElementCollection(Mp7JrsXPathParseContext * context, int & i, Mp7JrsNodeEnum & result)
{
	assert(context);
	
	i = -1; // Dont create an item
	
	if(context->LastNode != Mp7JrsNodePtr())
	{
		result.Insert(context->LastNode);
	}

	if(context->Create && result.Size() == 0)
	{
		switch(context->Operation)
		{
			case MP7JRS_XPATH_SELECT_ALL: // 
			case MP7JRS_XPATH_SELECT_TYPED_ALL: // [@xsi:type='bla']
			case MP7JRS_XPATH_SELECT_LAST: // [last()]
			case MP7JRS_XPATH_SELECT_LAST_TYPE: // [last() xsi:type] <----- FTT append if different type!
			case MP7JRS_XPATH_SELECT_TYPED_LAST: // [xsi:type last()]
			case MP7JRS_XPATH_FORCE_CREATE: // [position()>last()]
				if(context->UpperBoundary > -1 && context->CollectionSize >= context->UpperBoundary)
				{
					Mp7JrsXPathParseContext::ErrorBoundary(context);
				}
				i = context->CollectionSize; // Append
				return true;
				break;
			
			case MP7JRS_XPATH_SELECT_INDEX: // [i]
 			case MP7JRS_XPATH_SELECT_INDEXED_TYPE: // [i @xsi:type] <----- FTT insert or append if different type!
				if(context->UpperBoundary > -1 && context->CollectionSize >= context->UpperBoundary)
				{
					Mp7JrsXPathParseContext::ErrorBoundary(context);
				}
				if(context->StartIndex > context->CollectionSize)
				{
					Mp7JrsXPathParseContext::ErrorNoFillingGaps(context, context->CollectionSize);
				}
				i = context->StartIndex;
#if 0			
				if((context->ItemCount + 1) >= context->StartIndex)
				{
					if(context->UpperBoundary > -1 && context->CollectionSize >= context->UpperBoundary)
					{
						Mp7JrsXPathParseContext::ErrorBoundary(context);
					}
					i = context->Operation == Mp7Jrs_XPATH_SELECT_INDEXED_TYPE
						? context->LastValidItemIndex >= 0 ? context->LastValidItemIndex : context->CollectionSize // Insert or append
						: context->CollectionSize; // Append
					return true;
				}
				else
				{
					Mp7JrsXPathParseContext::ErrorNoFillingGaps(context, context->ItemCount);
				}
#endif				
				return true;
				break;

			case MP7JRS_XPATH_SELECT_TYPED_INDEX: // [@xsi:type i]
				if(context->UpperBoundary > -1 && context->CollectionSize >= context->UpperBoundary)
				{
					Mp7JrsXPathParseContext::ErrorBoundary(context);
				}
				if(context->StartIndex - context->ValidItemCount > 0)
				{
					Mp7JrsXPathParseContext::ErrorNoFillingGaps(context, context->ValidItemCount);
				}
				i = context->CollectionSize; // Append
				return true;
				break;
				
			// FTT: These are not used for Create!
			case MP7JRS_XPATH_SELECT_POSITION_RANGE: // [position()>A | position()<B]
			case MP7JRS_XPATH_SELECT_POSITION_RANGE_TYPE: // [position()>A | position()<B xsi:type]
			case MP7JRS_XPATH_SELECT_TYPED_POSITION_RANGE: // [xsi:type position()>A | position()<B]
			case MP7JRS_XPATH_SELECT_POSITION_LT: // [position()<A]
			case MP7JRS_XPATH_SELECT_POSITION_LT_TYPE: // [position()<A xsi:type]
			case MP7JRS_XPATH_SELECT_TYPED_POSITION_LT: // [xsi:type position()<A]
			case MP7JRS_XPATH_SELECT_POSITION_GT: // [position()>A]
			case MP7JRS_XPATH_SELECT_POSITION_GT_TYPE: // [position()>A xsi:type]
			case MP7JRS_XPATH_SELECT_TYPED_POSITION_GT: // [xsi:type position()>A]
				break;
			default:;			
		}
	}
	return false; // Don't create a new node
}

/* static */ bool Mp7JrsXPathParseContext::PostProcessSequenceCollection(Mp7JrsXPathParseContext * context, int & i, int & start, int & stop, Mp7JrsNodeEnum & result)
{
	assert(context);
	
	i = -1; // Dont create an additional sequence item
	start = stop = -1; // Dont shift elements
	
	if(context->LastNode != Mp7JrsNodePtr())
	{
		result.Insert(context->LastNode);
	}
	
	if(context->Create && result.Size() == 0)
	{
		switch(context->Operation)
		{
			case MP7JRS_XPATH_SELECT_ALL: // 
			case MP7JRS_XPATH_SELECT_TYPED_ALL: // [@xsi:type='bla']; Dont care about position
				if(context->FreeIndexList.size() == 0)
				{
					if(context->UpperBoundary > -1 && i >= context->UpperBoundary)
					{
						Mp7JrsXPathParseContext::ErrorBoundary(context);
					}
					i = context->CollectionSize; // Create and append another sequence item
				}
				else
				{
					start = stop = context->FreeIndexList.elementAt(0);
				}
				return true; // Something to do
				break;				
			
			case MP7JRS_XPATH_SELECT_LAST: // [last()]
			case MP7JRS_XPATH_SELECT_LAST_TYPE: // [last() xsi:type] <----- FTT append if different type!
			case MP7JRS_XPATH_SELECT_TYPED_LAST: // [xsi:type last()]
			case MP7JRS_XPATH_FORCE_CREATE: // [position()>last()]
				if(context->FreeIndexList.size() == 0)
				{
					if(context->UpperBoundary > -1 && i >= context->UpperBoundary)
					{
						Mp7JrsXPathParseContext::ErrorBoundary(context);
					}
					i = context->CollectionSize; // Create and append another sequence item
				}
				else
				{
					start = stop = context->FreeIndexList.elementAt(0);
				}				
				return true; // Something to do
				break;
	// TODO merge the two case sections!
	
			case MP7JRS_XPATH_SELECT_INDEX: // [i]
 			case MP7JRS_XPATH_SELECT_INDEXED_TYPE: // [i @xsi:type] <----- FTT insert or append if different type!
				if(context->FreeIndexList.size() == 0) // We have to create and append another sequence item.
				{
					if(context->UpperBoundary > -1 && i >= context->UpperBoundary)
					{
						Mp7JrsXPathParseContext::ErrorBoundary(context);
					}
					if(context->StartIndex - context->ItemCount > 0)
					{
						Mp7JrsXPathParseContext::ErrorNoFillingGaps(context, context->ItemCount);
					}
					i = context->StartIndex; // Insert or append another sequence item
				}
				else // Find gap and reorganise existing elements
				{
					if(context->StartIndex - context->ItemCount > 0)
					{
						Mp7JrsXPathParseContext::ErrorNoFillingGaps(context, context->ItemCount);
					}
					start = stop = context->FreeIndexList.elementAt(0);
					if(context->LastItemIndex > -1)
					{
						int j, k = -1, freeindex;
						for(j = 0; j < (int)context->FreeIndexList.size(); ++j)
						{
							freeindex = context->FreeIndexList.elementAt(j);
							if(freeindex < context->LastItemIndex)
							{
								k = freeindex;
							}
							else
							{
								if(k < 0)
								{
									k = freeindex;
								}
								break;
							}
						}
						start = k;
						stop = context->LastItemIndex;
					}
				}
				return true;
				break;
 			
			case MP7JRS_XPATH_SELECT_TYPED_INDEX: // [@xsi:type i]
				if(context->FreeIndexList.size() == 0) // We have to create and append another sequence item.
				{
					if(context->UpperBoundary > -1 && i >= context->UpperBoundary)
					{
						Mp7JrsXPathParseContext::ErrorBoundary(context);
					}
					if(context->StartIndex - context->ValidItemCount > 0)
					{
						Mp7JrsXPathParseContext::ErrorNoFillingGaps(context, context->ValidItemCount);
					}
					i = context->StartIndex; // Insert or append another choice item
				}
				else // Find gap and reorganise existing elements
				{
					if(context->StartIndex - context->ValidItemCount > 0)
					{
						Mp7JrsXPathParseContext::ErrorNoFillingGaps(context, context->ValidItemCount);
					}
					start = stop = context->FreeIndexList.elementAt(0);
					if(context->LastValidItemIndex > -1)
					{
						int j, k = -1, freeindex;
						for(j = 0; j < (int)context->FreeIndexList.size(); ++j)
						{
							freeindex = context->FreeIndexList.elementAt(j);
							if(freeindex < context->LastValidItemIndex)
							{
								k = freeindex;
							}
							else
							{
								if(k < 0)
								{
									k = freeindex;
								}
								break;
							}
						}
						start = k;
						stop = context->LastValidItemIndex;
					}
				}
				return true;
				break;
				
			// FTT: These are not used for Create!
			case MP7JRS_XPATH_SELECT_POSITION_RANGE: // [position()>A | position()<B]
			case MP7JRS_XPATH_SELECT_POSITION_RANGE_TYPE: // [position()>A | position()<B xsi:type]
			case MP7JRS_XPATH_SELECT_TYPED_POSITION_RANGE: // [xsi:type position()>A | position()<B]
			case MP7JRS_XPATH_SELECT_POSITION_LT: // [position()<A]
			case MP7JRS_XPATH_SELECT_POSITION_LT_TYPE: // [position()<A xsi:type]
			case MP7JRS_XPATH_SELECT_TYPED_POSITION_LT: // [xsi:type position()<A]
			case MP7JRS_XPATH_SELECT_POSITION_GT: // [position()>A]
			case MP7JRS_XPATH_SELECT_POSITION_GT_TYPE: // [position()>A xsi:type]
			case MP7JRS_XPATH_SELECT_TYPED_POSITION_GT: // [xsi:type position()>A]
				break;
			default:;			
		}
	}
	return false;
}

/* static */ bool Mp7JrsXPathParseContext::PostProcessChoiceCollection(Mp7JrsXPathParseContext * context, int & i, Mp7JrsNodeEnum & result)
{
	assert(context);
	
	i = -1; // Dont create an additional sequence item
	
	if(context->LastNode != Mp7JrsNodePtr())
	{
		result.Insert(context->LastNode);
	}
	
	if(context->Create && result.Size() == 0)
	{
		switch(context->Operation)
		{
			case MP7JRS_XPATH_SELECT_ALL: // 
			case MP7JRS_XPATH_SELECT_TYPED_ALL: // [@xsi:type='bla']; Dont care about position
				if(context->UpperBoundary > -1 && context->CollectionSize >= context->UpperBoundary)
				{
					Mp7JrsXPathParseContext::ErrorBoundary(context);
				}
				i = context->CollectionSize; // Create and append another sequence item
				return true; // Something to do
				break;				
			
			case MP7JRS_XPATH_SELECT_LAST: // [last()]
			case MP7JRS_XPATH_SELECT_LAST_TYPE: // [last() xsi:type] <----- FTT append if different type!
			case MP7JRS_XPATH_SELECT_TYPED_LAST: // [xsi:type last()]
			case MP7JRS_XPATH_FORCE_CREATE: // [position()>last()]
				if(context->UpperBoundary > -1 && context->CollectionSize >= context->UpperBoundary)
				{
					Mp7JrsXPathParseContext::ErrorBoundary(context);
				}
				i = context->CollectionSize; // Create and append another choice item
				return true; // Something to do
				break;
	// TODO merge the two case sections!
	
			case MP7JRS_XPATH_SELECT_INDEX: // [i]
 			case MP7JRS_XPATH_SELECT_INDEXED_TYPE: // [i @xsi:type] <----- FTT insert or append if different type!
				if(context->UpperBoundary > -1 && context->CollectionSize >= context->UpperBoundary)
				{
					Mp7JrsXPathParseContext::ErrorBoundary(context);
				}
				if(context->StartIndex > context->CollectionSize)
				{
					Mp7JrsXPathParseContext::ErrorNoFillingGaps(context, context->CollectionSize);
				}
				i = context->StartIndex; // Insert or append another choice item
				return true;
				break;
 			
			case MP7JRS_XPATH_SELECT_TYPED_INDEX: // [@xsi:type i]
				if(context->UpperBoundary > -1 && context->CollectionSize >= context->UpperBoundary)
				{
					Mp7JrsXPathParseContext::ErrorBoundary(context);
				}
				if(context->StartIndex > context->CollectionSize)
				{
					Mp7JrsXPathParseContext::ErrorNoFillingGaps(context, context->CollectionSize);
				}
				i = context->CollectionSize;
				return true;
				break;
				
			// FTT: These are not used for Create!
			case MP7JRS_XPATH_SELECT_POSITION_RANGE: // [position()>A | position()<B]
			case MP7JRS_XPATH_SELECT_POSITION_RANGE_TYPE: // [position()>A | position()<B xsi:type]
			case MP7JRS_XPATH_SELECT_TYPED_POSITION_RANGE: // [xsi:type position()>A | position()<B]
			case MP7JRS_XPATH_SELECT_POSITION_LT: // [position()<A]
			case MP7JRS_XPATH_SELECT_POSITION_LT_TYPE: // [position()<A xsi:type]
			case MP7JRS_XPATH_SELECT_TYPED_POSITION_LT: // [xsi:type position()<A]
			case MP7JRS_XPATH_SELECT_POSITION_GT: // [position()>A]
			case MP7JRS_XPATH_SELECT_POSITION_GT_TYPE: // [position()>A xsi:type]
			case MP7JRS_XPATH_SELECT_TYPED_POSITION_GT: // [xsi:type position()>A]
				break;
			default:;			
		}
	}
	return false;
}

/* static */ Mp7JrsNodeEnum Mp7JrsXPathParseContext::Root(Mp7JrsXPathParseContext * context, const Mp7JrsNodePtr node, Mp7JrsClientID client)
{
	assert(context);
	
	Mp7JrsNodeEnum result;
	const Mp7JrsNodePtr empty = Mp7JrsNodePtr();
	if(node == empty) return result;
	
	unsigned int rootid = Mp7JrsFactory::GetTypeIndex(X("urn:mpeg:mpeg7:schema:2004:Mpeg7_LocalType"));
	
	Mp7JrsNodePtr p = node;
	while(p != empty)
	{
		p = p->GetParent();
		if(p != empty && p->GetClassId() == rootid)
		{
			if(XMLString::stringLen(context->UnparsedXPath) > 0)
			{
				Mp7JrsXPathCreation flag = context->Create ? MP7JRS_FORCECREATE : MP7JRS_NOCREATE;
				result = p->GetOrCreateFromXPath(context->UnparsedXPath, client, flag, context);
			}
			else // Just "/"
			{
				context->Found = true;
				result.Insert(p);
			}
			break;
		}
	}
	return result;
}

/* static */ Mp7JrsNodeEnum Mp7JrsXPathParseContext::Up(Mp7JrsXPathParseContext * context, const Mp7JrsNodePtr node, Mp7JrsClientID client)
{
	/* FTT this source is copied verbatim from Node::UpwardWithXPath() */
	/* TODO: Replace string stuff with type ids */
	/* TODO: Refactor this */
	Mp7JrsNodeEnum result;
	const Mp7JrsNodePtr empty = Mp7JrsNodePtr();
	if(node == empty) return result;

	Mp7JrsNodePtr ancestor = node->GetParent();
	if(ancestor)
	{
		Mp7JrsNodeCollectionPtr ancestor_as_collection = ancestor;
		while (ancestor) 
		{
			// Look upward for a "meaningful" node to continue parsing.
			if (!ancestor_as_collection
				&& (!strstr(ancestor->GetTypeName(), "LocalType") // FTT This is bogus...
				// TODO: auf ContentName pruefen? -> mit Testcases abklaeren
				|| strstr(ancestor->GetTypeName(), "Mpeg7_LocalType"))
//				&& ancestor->GetContentName()
			)
				break;
			ancestor = ancestor->GetParent();
			ancestor_as_collection = ancestor;
		}
	}
	if(ancestor != empty)
	{
		if(XMLString::stringLen(context->UnparsedXPath) > 0)
		{
			Mp7JrsXPathCreation flag = context->Create ? MP7JRS_FORCECREATE : MP7JRS_NOCREATE;
			result = ancestor->GetOrCreateFromXPath(context->UnparsedXPath, client, flag, context);
		}
		else // Just ".."
		{
			context->Found = true;
			result.Insert(ancestor);
		}
	}
	return result;
}



/* static */ void Mp7JrsXPathParseContext::ErrorBoundary(Mp7JrsXPathParseContext * context)
{
	assert(context);
	
	char buf[32] = {0};
	XMLCh * msg = XMLString::transcode("XPath error:\n");
	Mp7JrsUtil::CatString(&msg, context->XPath);
	Mp7JrsUtil::CatString(&msg, X("\nCannot append/insert element. Upper boundary ("));
#if defined(XML_WIN32) || defined(WIN32) // Xerces 3.0 doesn't know XML_WIN32
	_snprintf_s(buf, 32, 31, "%d", context->UpperBoundary);
#elif defined(__APPLE__) || defined(APPLE) || defined(_APPLE) // Xerces 3.0 doesn't know XML_LINUX
	sprintf(buf, "%d", context->UpperBoundary);
#elif defined(XML_LINUX) || defined(LINUX) || defined(linux) // Xerces 3.0 doesn't know XML_LINUX
	sprintf(buf, "%d", context->UpperBoundary);
#else // XML_LINUX ...
#	error "TODO: Add platform specific headers for xpathparsecontext."
#endif		
	Mp7JrsUtil::CatString(&msg, X(buf));
	Mp7JrsUtil::CatString(&msg, X(") is reached."));
	Mp7JrsException e(MP7JRS_ERROR, msg);
	XMLString::release(&msg);
	throw e;
}

/* static */ void Mp7JrsXPathParseContext::ErrorUnallowedExpression(Mp7JrsXPathParseContext * context)
{
	assert(context);
	
	XMLCh * msg = XMLString::transcode("XPath error:\n");
	Mp7JrsUtil::CatString(&msg, context->XPath);
	Mp7JrsUtil::CatString(&msg, X("\nSorry, this expression is allowed only in Get() -without- OrCreate() calls."));
	Mp7JrsException e(MP7JRS_ERROR, msg);
	XMLString::release(&msg);
	throw e;			
}

/* static */ void Mp7JrsXPathParseContext::ErrorAbstractType(Mp7JrsXPathParseContext * context)
{
	assert(context);
	
	XMLCh * msg = XMLString::transcode("XPath error:\n");
	Mp7JrsUtil::CatString(&msg, context->XPath);
	Mp7JrsUtil::CatString(&msg, X("\nSorry, the type of \""));
	Mp7JrsUtil::CatString(&msg, context->ElementName);
	Mp7JrsUtil::CatString(&msg, X("\" is abstract and needs @xsi:type.\nCannot create this node."));
	Mp7JrsException e(MP7JRS_ERROR, msg);
	XMLString::release(&msg);
	throw e;
}

/* static */ void Mp7JrsXPathParseContext::ErrorInvalidIndex(Mp7JrsXPathParseContext * context)
{
	assert(context);
	
	char buf[32] = {0};
	XMLCh * msg = XMLString::transcode("XPath error:\n");
	Mp7JrsUtil::CatString(&msg, context->XPath);
	Mp7JrsUtil::CatString(&msg, X("\nSorry, invalid index "));
#if defined(XML_WIN32) || defined(WIN32) // Xerces 3.0 doesn't know XML_WIN32
	_snprintf_s(buf, 32, 31, "%d", context->StartIndex + 1); // Switch back to (1 based) XPath indexes 
#elif defined(__APPLE__) || defined(APPLE) || defined(_APPLE) // Xerces 3.0 doesn't know XML_LINUX
	sprintf(buf, "%d", context->StartIndex + 1); // Switch back to (1 based) XPath indexes 
#elif defined(XML_LINUX) || defined(LINUX) || defined(linux) // Xerces 3.0 doesn't know XML_LINUX
	sprintf(buf, "%d", context->StartIndex + 1); // Switch back to (1 based) XPath indexes 
#else // XML_LINUX ...
#	error "TODO: Add platform specific headers for xpathparsecontext."
#endif	
	Mp7JrsUtil::CatString(&msg, X(buf));
	Mp7JrsUtil::CatString(&msg, X(".\nCannot create node."));
	Mp7JrsException e(MP7JRS_ERROR, msg);
	throw e;
}

/* static */ void Mp7JrsXPathParseContext::ErrorInvalidEmptyItem(Mp7JrsXPathParseContext * context, int i)
{
	assert(context);
	
	char buf[32] = {0};
	XMLCh * msg = XMLString::transcode("XPath error:\n");
	Mp7JrsUtil::CatString(&msg, context->XPath);
	Mp7JrsUtil::CatString(&msg, X("\nEmpty node in collection at C++ index "));
#if defined(XML_WIN32) || defined(WIN32) // Xerces 3.0 doesn't know XML_WIN32
	_snprintf_s(buf, 32, 31, "%d", i); // We want clean collections with always non empty items inside!
#elif defined(__APPLE__) || defined(APPLE) || defined(_APPLE) // Xerces 3.0 doesn't know XML_LINUX
	sprintf(buf, "%d", i); // We want clean collections with always non empty items inside!
#elif defined(XML_LINUX) || defined(LINUX) || defined(linux) // Xerces 3.0 doesn't know XML_LINUX
	sprintf(buf, "%d", i); // We want clean collections with always non empty items inside!
#else // XML_LINUX ...
#	error "TODO: Add platform specific headers for xpathparsecontext."
#endif	
	Mp7JrsUtil::CatString(&msg, X(buf));
	Mp7JrsUtil::CatString(&msg, X("detected.\nIn order to keep collections clean, empty nodes are not allowed."));
	Mp7JrsException e(MP7JRS_ERROR, msg);
	throw e;
}

/* static */ void Mp7JrsXPathParseContext::ErrorNoFillingGaps(Mp7JrsXPathParseContext * context, unsigned int lastvalidindex)
{
	assert(context);
	
	char buf[32] = {0};
	XMLCh * msg = XMLString::transcode("XPath error:\n");
	Mp7JrsUtil::CatString(&msg, context->XPath);
	Mp7JrsUtil::CatString(&msg, X("\nSorry, XPath index "));
#if defined(XML_WIN32) || defined(WIN32) // Xerces 3.0 doesn't know XML_WIN32
	_snprintf_s(buf, 32, 31, "%d", context->StartIndex + 1); // Switch back to (1 based) XPath indexes 
#elif defined(__APPLE__) || defined(APPLE) || defined(_APPLE) // Xerces 3.0 doesn't know XML_LINUX
	sprintf(buf, "%d", context->StartIndex + 1); // Switch back to (1 based) XPath indexes 
#elif defined(XML_LINUX) || defined(LINUX) || defined(linux) // Xerces 3.0 doesn't know XML_LINUX
	sprintf(buf, "%d", context->StartIndex + 1); // Switch back to (1 based) XPath indexes 
#else // XML_LINUX ...
#	error "TODO: Add platform specific headers for xpathparsecontext."
#endif	
	Mp7JrsUtil::CatString(&msg, X(buf));
	Mp7JrsUtil::CatString(&msg, X(" is too high.\nCurrently there are "));
#if defined(XML_WIN32) || defined(WIN32) // Xerces 3.0 doesn't know XML_WIN32
	_snprintf_s(buf, 32, 31, "%u", lastvalidindex);
#elif defined(__APPLE__) || defined(APPLE) || defined(_APPLE) // Xerces 3.0 doesn't know XML_LINUX
	sprintf(buf, "%u", lastvalidindex);
#elif defined(XML_LINUX) || defined(LINUX) || defined(linux) // Xerces 3.0 doesn't know XML_LINUX
	sprintf(buf, "%u", lastvalidindex);
#else // XML_LINUX ...
#	error "TODO: Add platform specific headers for xpathparsecontext."
#endif	
	Mp7JrsUtil::CatString(&msg, X(buf));
	Mp7JrsUtil::CatString(&msg, X(" elements available for this request. Filling gaps is not allowed (only appending).\nCannot create this node."));
	// Give the user a change to correct his params...
	Mp7JrsMissingElementException e(MP7JRS_ERROR, msg, lastvalidindex);
	Mp7JrsXPathParseContext::Reset(context);
	throw e;
}

/* static */ void Mp7JrsXPathParseContext::ErrorCannotSetAttribute(Mp7JrsXPathParseContext * context)
{
	assert(context);
	
	XMLCh * msg = XMLString::transcode("XPath error:\n");
	Mp7JrsUtil::CatString(&msg, context->XPath);
	Mp7JrsUtil::CatString(&msg, X("\nSorry, cannot set attribute \""));
	Mp7JrsUtil::CatString(&msg, context->ElementName);
	if(context->XsiType)
	{
		Mp7JrsUtil::CatString(&msg, X("\" with type \""));
		Mp7JrsUtil::CatString(&msg, context->XsiType);
		Mp7JrsUtil::CatString(&msg, X("\"."));
	}
	else
	{
		Mp7JrsUtil::CatString(&msg, X("\" with default type."));
	}
	Mp7JrsException e(MP7JRS_ERROR, msg);
	Mp7JrsXPathParseContext::Reset(context);
	throw e;
}

/* static */ void Mp7JrsXPathParseContext::ErrorCannotSetElement(Mp7JrsXPathParseContext * context)
{
	assert(context);
	
	XMLCh * msg = XMLString::transcode("XPath error:\n");
	Mp7JrsUtil::CatString(&msg, context->XPath);
	Mp7JrsUtil::CatString(&msg, X("\nSorry, cannot set element \""));
	Mp7JrsUtil::CatString(&msg, context->ElementName);
	if(context->XsiType)
	{
		Mp7JrsUtil::CatString(&msg, X("\" with type \""));
		Mp7JrsUtil::CatString(&msg, context->XsiType);
		Mp7JrsUtil::CatString(&msg, X("\"."));
	}
	else
	{
		Mp7JrsUtil::CatString(&msg, X("\" with default type."));
	}
	Mp7JrsException e(MP7JRS_ERROR, msg);
	Mp7JrsXPathParseContext::Reset(context);
	throw e;
}

/* static */ void Mp7JrsXPathParseContext::ErrorUnknownSubElement(Mp7JrsXPathParseContext * context, const char * type)
{
	assert(context);
	
	XMLCh * msg = XMLString::transcode("XPath error:\n");
	Mp7JrsUtil::CatString(&msg, context->XPath);
	Mp7JrsUtil::CatString(&msg, X("\nSorry, \""));
	Mp7JrsUtil::CatString(&msg, X(type));
	Mp7JrsUtil::CatString(&msg, X("\" does not contain element \""));
	if(context->ElementName)
		Mp7JrsUtil::CatString(&msg, context->ElementName);
	Mp7JrsUtil::CatString(&msg, X("\" (or it is not of type \"Mp7JrsNodePtr\")."));
	Mp7JrsException e(MP7JRS_ERROR, msg);
	XMLString::release(&msg);
	throw e;
}

/* static */ void Mp7JrsXPathParseContext::ErrorNoSubElement(Mp7JrsXPathParseContext * context, const char * type)
{
	assert(context);
	
	XMLCh * msg = XMLString::transcode("XPath error:\n");
	Mp7JrsUtil::CatString(&msg, context->XPath);
	Mp7JrsUtil::CatString(&msg, X("\nSorry, \""));
	Mp7JrsUtil::CatString(&msg, X(type));
	Mp7JrsUtil::CatString(&msg, X("\" does not contain any element of type \"Mp7JrsNodePtr\"." ));
	Mp7JrsException e(MP7JRS_ERROR, msg);
	XMLString::release(&msg);
	throw e;
}

/* static */ void Mp7JrsXPathParseContext::ErrorUnsupportedType(Mp7JrsXPathParseContext * context, const char * type)
{
	assert(context);
	
	XMLCh * msg = XMLString::transcode("XPath error:\n");
	Mp7JrsUtil::CatString(&msg, context->XPath);
	Mp7JrsUtil::CatString(&msg, X("\nSorry, cannot deal with element of type \""));
	Mp7JrsUtil::CatString(&msg, X(type));
	Mp7JrsUtil::CatString(&msg, X("\"."));
	Mp7JrsException e(MP7JRS_ERROR, msg);
	XMLString::release(&msg);
	throw e;
}

/* static */ void Mp7JrsXPathParseContext::ErrorUnsupportedAttributeType(Mp7JrsXPathParseContext * context, const char * type)
{
	assert(context);
	
	XMLCh * msg = XMLString::transcode("XPath error:\n");
	Mp7JrsUtil::CatString(&msg, context->XPath);
	Mp7JrsUtil::CatString(&msg, X("\nSorry, XPATH expressions with simple attributes of type \""));
	Mp7JrsUtil::CatString(&msg, X(type));
	Mp7JrsUtil::CatString(&msg, X("\" are not supported yet."));
	Mp7JrsException e(MP7JRS_ERROR, msg);
	XMLString::release(&msg);
	throw e;
}

/* static */ void Mp7JrsXPathParseContext::ErrorWrongCaller(const XMLCh * xpath)
{
	XMLCh * msg = XMLString::transcode("XPath error:\n");
	if(xpath)
	{
		Mp7JrsUtil::CatString(&msg, xpath);
		Mp7JrsUtil::CatString(&msg, X("\n"));
	}
	Mp7JrsUtil::CatString(&msg, X("Sorry, calling class is not allowed to query XPaths (element classes only)!"));
	Mp7JrsException e(MP7JRS_ERROR, msg);
	XMLString::release(&msg);
	throw e;
}

/* static */ void Mp7JrsXPathParseContext::ErrorNotImplemented(const XMLCh * xpath, const char * type)
{
	XMLCh * msg = XMLString::transcode("XPath error:\n");
	if(xpath)
	{
		Mp7JrsUtil::CatString(&msg, xpath);
		Mp7JrsUtil::CatString(&msg, X("\n"));
	}
	Mp7JrsUtil::CatString(&msg, X("Sorry, XPath queries for class "));
	Mp7JrsUtil::CatString(&msg, X(type));
	Mp7JrsUtil::CatString(&msg, X(" is currently not supported."));
	Mp7JrsException e(MP7JRS_ERROR, msg);
	XMLString::release(&msg);
	throw e;
}


/* static */ bool Mp7JrsXPathParseContext::ContentFromString(Mp7JrsNode * node, const XMLCh * const txt)
{
	if(!(node && txt)) return false;
	
	XERCES_CPP_NAMESPACE_QUALIFIER DOMImplementation * impl;
	XERCES_CPP_NAMESPACE_QUALIFIER DOMDocument * doc;
	XERCES_CPP_NAMESPACE_QUALIFIER DOMElement * root;
	XERCES_CPP_NAMESPACE_QUALIFIER DOMElement * parent;
	XERCES_CPP_NAMESPACE_QUALIFIER DOMElement * current;
	XERCES_CPP_NAMESPACE_QUALIFIER DOMText * text;
	
	
	impl = XERCES_CPP_NAMESPACE_QUALIFIER DOMImplementationRegistry::getDOMImplementation(X("Core"));
	if(!impl) return false;
	
	try
	{
		doc = impl->createDocument(0		  // root element namespace URI.
			, X("dummy") // root element name
			, 0);	   // document type object (DTD).

		root = doc->getDocumentElement();
		parent = doc->createElement(X("parent"));
		root->appendChild(parent);
		text = doc->createTextNode(txt);
		parent->appendChild(text);
		bool res = node->Deserialize(doc, parent, &current);			
		doc->release();
		return res;
	}
	catch(...)
	{
	}			
	return false;
}

/* static */ XMLCh * Mp7JrsXPathParseContext::ContentToString(const Mp7JrsNode * node)
{
	XMLCh * res = (XMLCh*)NULL;
	if(!(node)) return res;
	
	XERCES_CPP_NAMESPACE_QUALIFIER DOMImplementation * impl;
	XERCES_CPP_NAMESPACE_QUALIFIER DOMDocument * doc;
	XERCES_CPP_NAMESPACE_QUALIFIER DOMElement * root;
	XERCES_CPP_NAMESPACE_QUALIFIER DOMElement * parent;
	XERCES_CPP_NAMESPACE_QUALIFIER DOMElement * newelement;
	
	
	impl = XERCES_CPP_NAMESPACE_QUALIFIER DOMImplementationRegistry::getDOMImplementation(X("Core"));
	if(!impl) return res;
	
	try
	{
		doc = impl->createDocument(0		  // root element namespace URI.
			, X("dummy") // root element name
			, 0);	   // document type object (DTD).

		root = doc->getDocumentElement();
		parent = doc->createElement(X("parent"));
		newelement = doc->createElement(X("element"));
		root->appendChild(parent);
		(const_cast< Mp7JrsNode* >(node))->Serialize(doc, parent, &newelement);
		res = Mp7JrsUtil::GetElementText(newelement);	// Replication is done within GetElementText()
		doc->release();
		return res;
	}
	catch(...)
	{
	}			
	return res;
}

/* static */  XMLCh * Mp7JrsXPathParseContext::ResolveTypeName(const XMLCh * xsitype)
{
  if(!xsitype) return NULL;

  XMLCh * result = NULL;
  
  XMLCh * prefix = Mp7JrsUtil::Prefix(xsitype); // Must be deallocated
  if(prefix && XMLString::stringLen(prefix) > 0) {
	// FTT NEW:
	const XMLCh * nsuri = Mp7JrsFactory::GetNamespaceUri(prefix);// Do not free this string!
	const XMLCh * localname = Mp7JrsUtil::LocalName(xsitype);  // Do not free this string!
	result = Mp7JrsUtil::QualifiedName(nsuri, localname);
  } else{ // Use default namespace uri
	result = Mp7JrsUtil::CatString(X("urn:mpeg:mpeg7:schema:2004:"), xsitype);
  }
  if(prefix) Mp7JrsUtil::deallocate(&prefix);	
  
  return result;
}
